local Fledder = {}	-- german: Flatterer (not really though...)
local lg = love.graphics

-- behavior:
-- sometimes they just sit around doing nothing
--> idea: when starting to fly / lay down set flying animation + change the body.z value over time
-- sometimes they fly around aimlessly
-- when threatened or ...(?) they go into aggressive mode
-- aggressive = attack nearby rock raiders
-- can be scared off by ultrasonic disturbances (player made at least)

function Fledder:create( data )
	local entity = {}
	
	entity.entityType = ENTITIES.fledder
	entity.entityCategory = ENTITYCATEGORY.monster
	
	entity.body = {
		shape = "circle",
		x = data.x or 0,
		y = data.y or 0,
		radius = 4,
		z = 64,	-- test: draw the entity above the origin of the body
	}
	
	entity.maxHealth = 150
	-- entity.health = entity.maxHealth
	
	-- 25/s = one mini tile per second
	entity.movementSpeed = 64
	
	entity.task = "idle"
	
	-- world position and mini tile index
	entity.movementDestination = {}-- x = 0, y = 0, mtx = 0, mty = 0 }
	
	entity.rendererLayer = "entities"
	
	-- animations
	entity.animations = {}
	entity.animations.idle = Animation.new( "animations/fledder/idle.lua" )
	entity.animations.idle:setRelativeOrigin( 0.5, 0.5 )
	
	-- set initial animation set
	entity.currentAnimation = "idle"
	
	-- when the unit recently lost or gained health, draw the healthbar for a few seconds
	entity.timeSinceLastHealthChange = 0
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		self.health = Health:create( self, self.maxHealth, false, 0 )	-- health component
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getPosition()
		return self.body.x, self.body.y
	end
	
	
	function entity:setPosition( x, y )
		self.body.x = x
		self.body.y = y
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- get the currently possible movement speed
	function entity:getCurrentMovementSpeed()
		local speed = self.movementSpeed
		
		-- TODO: modifiers?
		
		return speed
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- internal only
	-- mini tile coordinates and world position
	function entity:_setMovementDestination( mini_xi, mini_yi, x, y )
		self.movementDestination.x = x
		self.movementDestination.y = y
		self.movementDestination.mtx = mini_xi
		self.movementDestination.mty = mini_yi
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from extern: move to a specific location (from player move command)
	-- mini tile coordinates and world position
	function entity:moveToDestination( mini_xi, mini_yi, x, y )
		local ex, ey = self:getPosition()
		local start_txi, start_tyi = getTileIndex( ex, ey )
		local target_xi, target_yi = getTileIndex( x, y )
		
		-- extra: pre-check if there can be a path at all
		if not Pathfinding.getPath( start_txi, start_tyi, target_xi, target_yi, "flying_coarse" ) then
			return false
		end
		
		local start_xi, start_yi = getMiniTileIndex( ex, ey )
		local path = Pathfinding.getPath( start_xi, start_yi, mini_xi, mini_yi, "flying_fine" )
		if path then
			-- if canMove then
				
				-- new movement
				self:_setMovementDestination( mini_xi, mini_yi, x, y )
				self:setMovementPath( path )
				-- self:setTask( "moveTo" )
				return true
			-- end
			
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		
		-- behavior depending on the internal state
		self:behaviorTick( dt )
		
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:update( dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:behaviorTick( dt )
		
		local ex, ey = self:getPosition()
		local random = math.random
		
		if self.task == "idle" then
			
			-- TODO: only perform this check once per second
			if random() < 0.01 then
				self:setTask( "fly_around" )
			end
			
		elseif self.task == "fly_around" then
			
			if not self.movementPath then
				
				-- randomly move somewhere else
				local maxRange = 250
				-- note: a high maxRange might cause the entity to stray very far off...
				local x = ex + random(-maxRange, maxRange)
				local y = ey + random(-maxRange, maxRange)
				local mini_xi, mini_yi = getMiniTileIndex( x, y )
				self:moveToDestination( mini_xi, mini_yi, x, y )
				
			end
			
		end
		
		
		-- currently want to move somewhere
		if self.movementPath then
			local nSteps = #self.movementPath
			if nSteps > 0 then
				local start_xi, start_yi = getMiniTileIndex( ex, ey )
				if start_xi == self.movementPath[nSteps].x
				and start_yi == self.movementPath[nSteps].y then
					table.remove( self.movementPath, nSteps )
				else
					-- move to the next mini tile center
					local goal_x, goal_y = miniTileIndexToPosition( self.movementPath[nSteps].x, self.movementPath[nSteps].y )
					local direction = Tools.vectorSub( self.body, { x = goal_x, y = goal_y } )
					local vec2 = Tools.vectorSetLength( direction, self:getCurrentMovementSpeed() )
					self:move( vec2.x * dt, vec2.y * dt )
				end
			else
				-- already on the final mini tile
				-- move to the exact target location
				local direction = Tools.vectorSub( self.body, self.movementDestination )
				local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
				self:move( vec2.x, vec2.y )
				
				if math.abs( self.body.x - self.movementDestination.x ) < 2
				and math.abs( self.body.y - self.movementDestination.y ) < 2 then
					self.movementPath = nil
					self:onMovementDestinationReached()
				end
			end
		end
			
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:onMovementDestinationReached()
		
		--self:setTask( "idle" )
		
		local ex, ey = self:getPosition()
		local random = math.random
		
		if self.task == "fly_around" then
			local tx, ty = getTileIndex( ex, ey )
			if Game.terrain:getTerrainTypeAtTile( tx, ty ) == 1 and random() < 0.2 then
				-- TODO: land...
				self:setTask( "idle" )
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setMovementPath( path )
		self.movementPath = path
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:move( dx, dy )
		self.body.x = self.body.x + (dx or 0)
		self.body.y = self.body.y + (dy or 0)
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set animation etc.
	function entity:setTask( task )
		
		-- already doing that?
		if self.task == task then
			return
		end
		
		-- update task
		self.task = task
		-- print( "fledder: set task:", task )
		
		if task == "idle" then
			self.currentAnimation = "idle"
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from game when a hurt event occured
	function entity:onHurt( damage )
		
		-- no modifications etc. needed here, simply eat the damage
		-- self.health = self.health - damage
		-- self:onHealthChanged()
	end
	
	
	function entity:onHealthChanged()
		self.timeSinceLastHealthChange = 0
		-- print( "fledder:onHealthChanged" )
		-- print( self.health:getHealth() )
		if self.health:getHealth() <= 0 then
			self:die()
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:die()
		-- TODO: animation, sound ...
		EntityManager:remove( self )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:setPosition( x, y - body.z )	-- extra: offset makes it look like the entity is in the air
			anim:draw()
		end
		
		-- "shadow"
		lg.setColor( 0.1, 0.1, 0.1, 0.5 )
		lg.ellipse( "fill", x, y, 6, 3 )
		
		-- healthbar
		if Game.drawEntityStatusOverhead
		or self.timeSinceLastHealthChange < 4 then
			local hb_y_offset = body.z + 24	-- draw above the entity
			self.health:draw( x, y - hb_y_offset )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.health:remove()
		self = nil
	end
	
	
	return entity
end


return Fledder
